@@ -7,7 +7,8 @@ from hashedindex.textparser import (
     NullStemmer,
 )
 from hashedixsearch._internal import (
-    _longest_prefix,
+    _candidate_matches,
+    _is_separator,
     _ngram_to_term,
     SynonymAnalyzer,
 )
@@ -127,7 +128,7 @@ def highlight(query, terms, stemmer=None, synonyms=None, case_sensitive=True):
 
     # If no terms are provided to match on, do not attempt highlighting
     if not terms:
-        return query
+        return escape(query)
 
     terms = {term: list(term) for term in terms}
     max_n = max(len(term) for term in terms.values())
@@ -148,7 +149,7 @@ def highlight(query, terms, stemmer=None, synonyms=None, case_sensitive=True):
 
     # If we did not generate any ngrams, do not attempt highlighting
     if not ngrams:
-        return query
+        return escape(query)
 
     # Tail the ngram list with ngrams of decreasing length
     final_ngram = ngrams[-1]
@@ -158,6 +159,8 @@ def highlight(query, terms, stemmer=None, synonyms=None, case_sensitive=True):
     # Build up a marked-up representation of the original query
     tag = None
     markup = ""
+    accumulator = ""
+
     for ngram in ngrams:
 
         # Stop when we reach an empty end-of-stream ngram
@@ -166,21 +169,33 @@ def highlight(query, terms, stemmer=None, synonyms=None, case_sensitive=True):
 
         # Determine whether any of the highlighting terms match
         ngram_term = _ngram_to_term(ngram, stemmer, case_sensitive)
-        longest_term = _longest_prefix(ngram_term, terms)
 
-        # Begin markup if a prefix match was found
-        if longest_term and longest_term != tag:
-            markup += "<mark>"
-            tag = longest_term
+        if not tag and not _is_separator(ngram_term[0]):
+            tag = _candidate_matches(ngram_term, terms)
+            markup += accumulator
+            accumulator = ""
 
         # Consume one token at a time
-        markup += escape(ngram[0])
+        accumulator += escape(ngram[0])
+        emit = tag is None
 
-        # Close markup when all of a tag's tokens are consumed
-        if tag and tag[0] == ngram_term[0]:
-            tag.pop(0)
-            if not tag:
-                markup += "</mark>"
+        # Advance the match window of each candidate tag element
+        if tag and not _is_separator(ngram_term[0]):
+            tag = {
+                term: tokens[1:]
+                for term, tokens in tag.items()
+                if tokens[0] == ngram_term[0]
+            }
+
+            # Close the markup when any term's tokens have been consumed
+            if not all(tag.values()):
+                accumulator = f"<mark>{accumulator}</mark>"
+                emit = True
                 tag = None
 
-    return markup
+        # Output accumulated tokens when we encounter emit-points
+        if emit:
+            markup += accumulator
+            accumulator = ""
+
+    return markup + accumulator
